{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
--module SqrParser(pSqrProgram, showResult) where
module SqrParser where
import Text.Parsec hiding (space, spaces)
import Text.Parsec.Pos
import Text.Parsec.Error

import Control.Applicative ((<*>), (*>), (<*), (<$>), liftA2)
import Control.Monad
import Control.Monad.Trans
import Data.Char
import qualified Data.Map as M
import Data.List
import Data.Maybe
import Data.Text (pack, unpack, strip)
import Debug.Trace
import SqrppParser
import SqrFile

run p s = runParserT p initialState "" s >>= either (print) (print)

runExpr s = runParserT pExpr initialState "" s >>= either (print) (print)

trim :: String -> String
trim = unpack . strip . pack

pTrailingHyphen = do
    try (char '-' >> eol)
    setContinued True
    return ()

setContinued b = do
    st <- getState
    if b == continued st then return () else (setState st {continued=b})

{-
commandable = return ()
-}

space = (nbsp >> return ()) <|> pTrailingHyphen <|> (eol >> return ())

spaces :: SqrParser ()
spaces = skipMany space

lexeme p = do r <- p
              spaces
              return r



parens p = lexeme (char '(') *> (lexeme p) <* lexeme (char ')')

brackets p = (lexeme (char '[')) *> (lexeme p) <* (lexeme (char ']'))

pSqrProgram :: SqrParser ParserState
pSqrProgram = do
      setBuiltinVariables
      skipMany space
      lexeme (many pSection)
      eof
      getState

pSection =  pSetup
        <|> pProcedure
        <|> pProgram
        <|> pReport
        <|> pHeading
        <|> pFooting

section :: String -> SqrParser () -> SqrParser ()
section keyword tokenParser = lexeme $ do
    command ("BEGIN-" ++ keyword)
    dbgPrint $ ">> BEGIN-" ++ keyword
    manyTill tokenParser (pEndLine ("END-" ++ keyword))
    return ()

pSetup :: SqrParser ()
pSetup = section "SETUP" $
              pDeclareVariable Global
          <|> pOtherDeclare
          <|> pAnyParagraph
          <|> pAnyStatement

pProgram :: SqrParser ()
pProgram = section "PROGRAM" (pAnyParagraph <|> pAnyStatement)

pReport :: SqrParser ()
pReport = section "REPORT" (pAnyParagraph <|> pAnyStatement)

pHeading :: SqrParser ()
pHeading = lexeme $ do
    command "BEGIN-HEADING"
    dbgPrint $ ">> BEGIN-HEADING"
    lexeme (many1 digit)
    manyTill (pAnyParagraph <|> pAnyStatement) (pEndLine "END-HEADING")
    return ()

pFooting :: SqrParser ()
pFooting = lexeme $ do
    command "BEGIN-FOOTING"
    dbgPrint $ ">> BEGIN-FOOTING"
    lexeme (many1 digit)
    manyTill (pAnyParagraph <|> pAnyStatement) (pEndLine "END-FOOTING")
    return ()





pProcedure :: SqrParser ()
pProcedure = try $ lexeme $ do
    command "BEGIN-PROCEDURE"
    pos <- getPosition
    name <- pProcedureName

    (isLocal, params) <- choice [ (parens paramList >>= (return . ((,) True)))
                              , (symbol "LOCAL" >> return (True, []))
                              , return (False, [])
                              ]
    dbgPrint $ ">> BEGIN-PROCEDURE : " ++ name ++ (if isLocal then " LOCAL" else "")

    let scope = if isLocal then (Local (fmap toUpper name)) else Global
    setCurrentScope scope
    ret <- ln (return $ name ++ "\t(" ++ (show scope) ++ ")")

    let proc = Procedure { procName=name
                         , procScope=scope
                         , procPosition=pos
                         , parameters=(fmap fst params) }
    saveProcedure proc

    if isLocal then optional (pDeclareVariable scope)
      else optional (command "DECLARE-VARIABLE" >> pFail "(SQR 7207) This command is only allowed within local procedures.")

    scope <- getCurrentScope
    Data.List.foldl (>>) (return ()) (fmap (\(param, (var_name, pos)) -> saveVariable scope (varSymbol $ varIdentifier param) var_name False pos) params)

    spaces
    -- ***** --
    manyTill (invalidDeclVar <|> try pAnyParagraph <|> try pAnyStatement) (pEndLine "END-PROCEDURE")
    -- ***** --


    setCurrentScope Global

--    return ret
    return ()
  where paramList  = option [] (parameters <* (optional (lexeme $ try $ (char ',' >> skipMany nbsp))))
        parameters = try $ (:) <$> parameter <*> (option [] (try (separator >> parameters)))
        parameter  = try $ do
                          { calltype  <- option CallByValue (lexeme (char ':') >> return CallByReference)
                          ; pos       <- getPosition
                          ; vsymbol   <- pAnyVarSymbol
                          ; var_name  <- pParameterVarName
                          ; skipMany nbsp
                          ; return (Parameter (createVarIdentifier vsymbol var_name) calltype, (var_name, pos))
                          }
        separator = lexeme $ try $ (char ',' >> return ()) <|> (eol >> return ())
        invalidDeclVar = command "DECLARE-VARIABLE" >> pFail "(SQR 7208) This command must be before all other commands in the procedure."

pParameterVarName = many1 (noneOf (chr 0x1A : " \t\n\r),!") <|> try (char '!' >> char '!'))


pProcedureName = lexeme $ many1 (noneOf excepts <|> (char '!' >> char '!'))
  where excepts = chr 26 : " \t\n\r('!"



pOtherDeclare :: SqrParser ()
pOtherDeclare = lexeme $
                    (   decl "CHART"
                    <|> decl "COLOR-MAP"
                    <|> decl "CONNECTION"
                    <|> decl "IMAGE"
                    <|> decl "LAYOUT"
                    <|> decl "PRINTER"
                    <|> decl "PROCEDURE"
                    <|> decl "REPORT"
                    <|> decl "TOC"
                    )
  where decl s = do
              try $ command ("DECLARE-" ++ s)
              spaces
              dbgPrint $ ">> " ++ "DECLARE-" ++ s
              manyTill (try pAnyParagraph <|> try pAnyStatement) (pEndLine "END-DECLARE")
              return ()

pDeclareVariable :: Scope -> SqrParser ()
pDeclareVariable scope = lexeme $ do
    command "DECLARE-VARIABLE"
    dbgPrint ">> DECLARE-VARIABLE"
    manyTill pDeclVarQualifier (pEndLine "END-DECLARE")
    return ()


pDeclVarQualifier = pDefaultNumeric
                <|> pDeclDecimalVars
                <|> pDeclIntegerVars
                <|> pDeclFloatVars
                <|> pDeclTextVars
                <|> pDeclDateVars
                <|> noQualifierError
                <|> qualifierError


noQualifierError = do oneOf "#$"
                      many1 (noneOf " \t\n\r") <|> pFail "(SQR 7210) Invalid Variable name specified."
                      pFail "(SQR 7206) The Variable type has not been specified."

qualifierError = do
    tok <- many1 (noneOf " \t\n\r")
    pFail $ "(SQR 5303) Invalid qualifier '" ++ tok
            ++ "'.  Valid qualifiers are: "
            ++ "default-numeric  decimal  integer  float  text  date  end-declare"


pDefaultNumeric = lexeme $ try $ do
    symbol "DEFAULT-NUMERIC"
    lexeme (char '=')
    var_type <- pInteger <|>  pDecimal <|> pFloat
    return ()


pInteger = symbol "INTEGER" >> return INTEGER


pFloat   = symbol "FLOAT"   >> return FLOAT


pDecimal = lexeme $ do
      symbol "DECIMAL"
      default_prec <- defaultPrecision
      precision <- option default_prec pPrecision
      return $ DECIMAL precision


pDeclDecimalVars = pDecimal >>= pDecimalVarList >> return ()


pDecimalVarList = (\(DECIMAL prec) -> many1 (pDecimalVarSpec prec))


pDecimalVarSpec default_prec = lexeme $ do
      (vname, ubar, pos) <- lexeme $ pDeclVarName Sharp
      precision <- option default_prec pPrecision
      let vtype = DECIMAL precision
      defineVariable vtype vname ubar pos

defaultPrecision = return 5

pPrecision = do
    n <- parens (many1 digit)
    return $ read n

pDeclIntegerVars = lexeme $ pInteger >>= pDeclVars Sharp

pDeclFloatVars   = lexeme $ pFloat   >>= pDeclVars Sharp

pDeclTextVars = lexeme $ do
        symbol "TEXT"
        pDeclVars Dollar TEXT

pDeclDateVars = lexeme $ do
        symbol "DATE"
        pDeclVars Dollar DATE

pDeclVars var_symbol var_type = lexeme $ try $ do
    vlist <- pVarList var_symbol
    foldl1 (>>) (fmap (\(var_symbol, var_name, ubar, pos) -> defineVariable var_type var_name ubar pos) vlist)
    return ()

pVarList var_symbol = do
    var_name_list <- many1 (lexeme $ pDeclVarName var_symbol <* follewedBySpace)
    return $ fmap (\(vname, ubar, pos) -> (var_symbol, vname, ubar, pos)  ) var_name_list

--                        !!! Never call lexeme here !!!
pDeclVarName var_symbol = do
          pos <- getPosition
          varSymb <- pAnyVarSymbol
          if (varSymb /= var_symbol)
              then case var_symbol of
                  Sharp -> pFail "(SQR 7204) A numeric variable name is required here."
                  Dollar -> pFail "(SQR 7203) A string variable name is required here."
              else return ()
          underbar <- option "" (string "_")
          var_name <- var_name_prsr
          scope <- getCurrentScope
          let ubar = (underbar == "_") && (scope == Global)
          return (underbar ++ var_name, ubar, pos)

  where var_name_prsr = (many1 var_name_char) <|> pFail "(SQR 7210) Invalid Variable name specified."
        var_name_char = oneOf (')':'-':['0'..'9'] ++ ['A'..'Z'] ++ "_" ++ ['a'..'z'] ++ [(chr 0xA6).. (chr 0xdf)])


ln p = do
  pos <- getPosition
  let flname = sourceName pos
      lineno = sourceLine pos
  body <- p
  return $ flname ++ "\t" ++ (show lineno) ++ "\t" ++ body ++ "\n"


rowNo :: SqrParser String
rowNo = do
    seqNo <- sequenceNo <$> getState
    pos <- getPosition
    return $ (show seqNo) ++ "\t"
          ++ (sourceName pos) ++ "\t"
          ++ (show (sourceLine pos)) ++ "\t"


pString :: SqrParser String
pString = do
      char '\''
      s <- (many (noneOf "\'\n\r" <|> quote <|> try (eol >> return ' ')) <* char '\'') <|> pFail "(SQR 3501) Did not find end of literal."
      return $ '\'' :( s ++ "\'" )
  where quote = try (char '\'' >> char '\'')

{-
pNumVar :: SqrParser String
pNumVar = do
      char '#'
      var <- many (noneOf (chr 26:" \t\n\r")) <|> pFail "(SQR 3501) Did not find end of literal."
      return ('#': var)
-}
{-
pNumVar :: SqrParser String
pNumVar = do
      char '#'
      varName <- many (noneOf (chr 26:" \t\n\r,()*=")) <|> pFail "(SQR 3501) Did not find end of literal."
      return ('#': varName)
-}


pUnexpectedEndCommand = try $ do
      stringCI "END-"
      choice (fmap command keywords)
      pFail "(SQR 4702)ff Line found outside paragraph."
      return ()
  where keywords =
          [ "FOOTING"
          , "HEADING"
          , "PROCEDURE"
          , "PROGRAM"
          , "REPORT"
          , "SETUP"
          , "DECLARE"
          ]

-- pAnyTokenList = join <$> manyTill anyTok (lookAhead $ pCommand (choice commandList))

pAnyParagraph = lexeme $ try $ (paragraph "SQL" <|> paragraph "SELECT")
  where paragraph s = do
                    command ("BEGIN-" ++ s)
                    spaces
                    dbgPrint $ ">> BEGIN-" ++ s
                    manyTill (try pAnyStatement) (pEndLine $ "END-" ++ s)
                    return ()

pEndLine s = lexeme $ try $ do
    spaces
    stringCI s
    follewedBySpace
    dbgPrint $ "<< " ++ s
    many (noneOf "\n\r")
    return ()

pAnyStatement = lexeme $ try $ do
          skipMany nbsp
          pUnexpectedEndCommand
              <|> pDoStatement
--          <|> try (pLetStatement)
--          <|> try (pIfStatement)
--          <|> try (pWhileStatement)
              <|> (many1 (anyTok <* skipMany ( (nbsp >> return ()) <|> pTrailingHyphen))  >> return ())
          return ()
-- >>= (\xs -> (return $ (concat xs) ++"\n")) >>= dbgPrint

anyTok =  do
      r <- (pString
--            <|> try (pSqlBindVar)
            <|> try (pAnyVariable var_name_parser)
            <|> try (count 1 (oneOf opChars))
            <|> try (anyWord))
      return r
  where var_name_parser = (:) <$> noneOf (chr 0x1A : " \t\n\r") <*> many (noneOf (chr 0x1A : " \t\n\r()[],!|+^*/%<>="))

opChars = "()[],!|+^*/%<>="

pSqlBindVar = brackets (pAnyVariable var_name_parser )
  where var_name_parser = many1 (noneOf excepts)
        excepts = " \t\n\r[]"
{-
pLooseVarName = many1 (noneOf excepts)
  where excepts = " \t\n\r"
-}

pDoStatement = lexeme $ try $ do
    command "DO"
    nm <- lexeme procName
    args <- option ([] ::[String]) $ parens argList
    dbgPrint $ "\tDo " ++ nm ++ "(" ++ (showListItems args) ++ ")"
    return ()
  where procName = many1 (noneOf (chr 0x1A : " \t\n\r(!\'"))

argList  = option ([] ::[String]) (arguments <* (optional (lexeme $ try $ (char ',' >> skipMany nbsp))))
  where arguments = try $ (:) <$> argument <*> (option ([] ::[String]) (try (separator >> arguments)))
        argument  = try $ (pNumLiteral <|> pTextLiteral <|> (pAnyVariable pParameterVarName)) <* skipMany nbsp
        separator = lexeme $ try $ (char ',' >> return ()) <|> (eol >> return ())

pLetStatement = lexeme $ do
    command "Let"
    var <- lvalue
    lexeme (char '=')
    expr <- pExpr
    return $ "LET [" ++ var ++ "] :=" ++ expr
  where lvalue = pVar <|> pArrayElement


pIfStatement = lexeme $ do
    command "if"
    dbgPrint "*** IF ***"
    expr <- lexeme $ pExpr
    dbgPrint expr
    body1
  where body1 = try end_if <|> try (els >> body2) <|> (pAnyStatement >> body1)
        body2 = try end_if <|> try (pAnyStatement >> body2)
        els = pOnlyCommandLine "else"
        end_if = pOnlyCommandLine "end-if"

pOnlyCommandLine keyword = lexeme $ try $ do
    stringCI keyword
    skipMany nbsp
    (eol >> return ()) <|> eof
    dbgPrint $ "<** " ++ keyword ++ " **>"
    return ""

pWhileStatement = lexeme $ do
    command "while"
    pExpr
    body
  where body = end_while <|> (pAnyStatement >> body)
        end_while = pOnlyCommandLine "end-while"

command :: String -> SqrParser String
command s = pCommand (stringCI s)

pCommand p = lexeme $ try $ do
    spaces
    p <* follewedBySpace

anyWord = (:) <$> (noneOf ("#$" ++ excepts)) <*> many (noneOf excepts)
  where excepts = " \t\n\r\'" ++ opChars



follewedBySpace = lookAhead $ try $ (satisfy (\c -> isSpace c) >> return ()) <|> eof

symbol :: String -> SqrParser String
symbol s = lexeme $ try $ stringCI s <* pEndSymbol

pEndSymbol = lookAhead $ try $ (    satisfy (\c -> isSpace c)
                                <|> oneOf symbols
                                <|> (eof >> return ' ') )
    where symbols = "!|(),+^*/%<>="


setBuiltinVariables = getVarMap Global
    >>= builtin INTEGER "#current-column"
    >>= builtin INTEGER "#current-line"
    >>= builtin INTEGER "#end-file"
    >>= builtin INTEGER "#page-count"
    >>= builtin INTEGER "#return-status"
    >>= builtin INTEGER "#sql-count"
    >>= builtin INTEGER "#sql-status"
    >>= builtin INTEGER "#sqr-max-columns"
    >>= builtin INTEGER "#sqr-max-lines"
    >>= builtin INTEGER "#sqr-pid"
    >>= builtin DATE    "$current-date"
    >>= builtin TEXT    "$sql-error"
    >>= builtin TEXT    "$sqr-database"
    >>= builtin TEXT    "$sqr-dbcs"
    >>= builtin TEXT    "$sqr-encoding"
    >>= builtin TEXT    "$sqr-hostname"
    >>= builtin TEXT    "$sqr-locale"
    >>= builtin TEXT    "$sqr-platform"
    >>= builtin TEXT    "$sqr-program"
    >>= builtin TEXT    "$sqr-report"
    >>= builtin TEXT    "$sqr-ver"
    >>= builtin TEXT    "$username"
    >>= updateVarMap Global
  where builtin var_type var_text = setVariable (typeToVarSymbol var_type) (tail var_text) False nullPosition (updateAttr var_type)
        nullPosition = initialPos ""
        updateAttr var_type = (\attr -> attr {varType=var_type, defined=True, usedCount=9999})


createVarIdentifier var_symbol var_name = VarIdentifier { varName=(fmap toUpper var_name)
                                                        , varSymbol=var_symbol
                                                        }
typeToVarSymbol :: VarType -> VarSymbol
typeToVarSymbol typ = case typ of
                         TEXT -> Dollar
                         DATE -> Dollar
                         _    -> Sharp


defineVariable :: VarType -> VarName -> StartWithUnderbar -> SourcePos -> SqrParser ()
defineVariable var_type var_name ubar pos = do
    scope <- getCurrentScope
    oldVMap <- getVarMap scope
    newVMap <- setVariable (typeToVarSymbol var_type) var_name ubar pos updateAttr oldVMap
    updateVarMap scope newVMap
  where updateAttr = (\attr -> attr {varType=var_type, defined=True, usedCount=0})

saveVariable :: Scope -> VarSymbol -> VarName -> StartWithUnderbar -> SourcePos -> SqrParser ()
saveVariable scope var_symbol var_name ubar pos = do
    st <- getState
    oldVMap <- getVarMap scope
    newVMap <- setVariable var_symbol var_name ubar pos id oldVMap
    updateVarMap scope newVMap


updateVarMap scope vmap = do
    st <- getState
    let newScopeMap = M.insert scope vmap (scopeMap st)
    setState st {scopeMap=newScopeMap}


setVariable :: VarSymbol -> VarName -> StartWithUnderbar -> SourcePos -> (VarAttr -> VarAttr) -> VarMap -> SqrParser VarMap
setVariable var_symbol var_name ubar pos updateAttr varMap = do
    repPos <- getReportPos pos
    let var = VarIdentifier {varName=(fmap toUpper var_name),varSymbol=var_symbol}
        newAttr = updateAttr $ updateDefault $ maybe (initialVarAttr var) id (M.lookup var varMap)
        initialVarAttr var = VarAttr  { varType= if (varSymbol var == Sharp) then TEXT else defaultNumericVarType
                                      , caseSensitiveName=var_name
                                      , defined=False
                                      , varPosition=pos
                                      , reportPos=repPos
                                      , usedCount=0
                                      , warnUnderbarPos=M.empty
                                      }
        updateDefault attr = let rep_pos = maybe repPos Just (reportPos attr)
                                in attr { usedCount=(1 + usedCount attr)
                                        , reportPos =  rep_pos
                                        , warnUnderbarPos = let pos_list = (warnUnderbarPos attr)
                                                              in case repPos of
                                                                  Just pos -> let fline = fileLine pos in
                                                                                  (if ubar && not (M.member fline pos_list)
                                                                                      then M.insert fline pos pos_list
                                                                                      else pos_list)
                                                                  Nothing -> pos_list
                                        , caseSensitiveName=if (reportPos attr) == Nothing && repPos /= Nothing
                                                                    then var_name
                                                                    else (caseSensitiveName attr)
                                        }
    return $ M.insert var newAttr varMap


defaultNumericVarType = INTEGER

getReportPos pos = do
    st <- getState
    return $
      let func = (\f -> f == (rootSourceName st) || ((recursiveCheck st) && matchSourceNameForReport st f)) in
          if func (sourceName pos) then Just pos else Nothing

setCurrentScope scope = do
    st <- getState
    setState st {currentScope=scope}

getCurrentScope = getState >>= return . currentScope

getVarMap scope = do
    st <- getState
    return $ maybe M.empty id (M.lookup scope (scopeMap st))


saveProcedure :: Procedure -> SqrParser ()
saveProcedure proc = do
    st <- getState
    let oldProcMap = procMap st
        key = fmap toUpper (procName proc)
        newProcMap = M.insert key proc oldProcMap
    setState (st {procMap=newProcMap})


pVarOrNumLiteral = pNumVar <|> pNumLiteral

pNumLiteral = naturalOrFloat >>= return . show


pHexDecimal = lexeme $ try $ do
    stringCI "0x"
    many1 digit



pTextLiteral = lexeme $ try $ do
      s <- (char '\'' *> many (noneOf "\'\n\r" <|> quote) <* char '\'') <|> pFail "(SQR 3501) Did not find end of literal."
      return $ "'" ++ s ++ "'"
  where quote = try (char '\'' >> char '\'')




pExpr = lexeme $ try $ pTerm +++ (option "" (pBinaryOp +++ pExpr))

pTerm = lexeme $ try $ (try pLiteral)
                    <|> try (pUnaryOp +++ pExpr)
                    <|> (try pVar)
                    <|> try (parens pExpr)
                    <|> pFunctionCall
                    <|> pArrayElement
                    <|> pColumn

pLiteral = try $ try pNumLiteral <|> try pTextLiteral

pVar = try $ pAnyVariable pStrictName

pColumn = lexeme $ try $ do
    symbol <- char '&'
    col_name <- colname
    return $ symbol : col_name
  where colname = many1 $ oneOf ("." ++ ['0'..'9'] ++ ['A'..'Z'] ++ ['a'..'z'] ++ "_" ++ [(chr 0xA6).. (chr 0xdf)])

pAnyVariable varNameParser = pVariable pAnyVarSymbol varNameParser

pNumVar = try $ pVariable pNumVarSymbol pStrictName

pVariable symbolParser varNameParser = try $ do
    pos           <- getPosition
    vsymbol       <- symbolParser
    underbar      <- option "" (string "_")
    varName       <- varNameParser
    (scope, ubar) <- varScope underbar
    saveVariable scope vsymbol ((if ubar then  "_" else "") ++ varName) ubar pos
    return $ (show vsymbol) ++ ( underbar ++ varName )

pAnyVarSymbol = pNumVarSymbol <|> pTextVarSymbol
pNumVarSymbol  = char '#' >> return Sharp
pTextVarSymbol = char '$' >> return Dollar

pStrictName = (:) <$> varNameChar <*> (join <$> (many (option "" (string "-") +++ (many1 varNameChar))))
  where varNameChar = oneOf (['0'..'9'] ++ ['A'..'Z'] ++ ['a'..'z'] ++ "_" ++ [(chr 0xA6).. (chr 0xDF)])


varScope underbar = do
    curScope <- getCurrentScope
    let ubar = (underbar == "_")
    return $ case curScope of
              Global  -> ( Global, ubar)
              Local _ -> ( (if ubar then Global else curScope), False)

pArrayElement = lexeme $ try $ do
    arrayName <- pStrictName
    char '.'
    fieldName <- pStrictName
    indexes <- parens pArgs
    return $ arrayName ++ "." ++ fieldName ++ "(" ++ (showListItems indexes) ++ ")"

pUnaryOp = pOperator "-" <|> pOperator "+" <|> pOperator "Not"

pFunctionCall = lexeme $ try $ do
    funcname <- lexeme pFuncName
    args <- parens pArgs
    return $ funcname ++ "(" ++ (showListItems args) ++ ")"
  where pFuncName = many1 (letter <|> digit <|> char '_')

pArgs = lexeme $ try $ sepBy pExpr (lexeme $ char ',')

showListItems = concat . (intersperse ", ")

pOperator = lexeme . try . stringCI

pBinaryOp = do
  op <- lexeme $ pArithmeticOp <|> pRelOp <|> pTextOp <|> pBitOp
  return $ " <" ++ op ++ "> "

pArithmeticOp = lexeme (
        op "+"  (+)
    <|> op "-"  (-)
    <|> op "*"  (*)
    <|> op "/"  (div)
    <|> op "%"  (mod)
    <|> op "^"  (^)
    )
    >>= return .fst
  where op s o = try (string s *> return (s, o))

pRelOp = lexeme (
        op "="   (==)
    <|> op "!!=" (/=)
    <|> op "!="  (/=)
    <|> op "<>"  (/=)
    <|> op "<="  (<=)
    <|> op ">="  (>=)
    <|> op "<"   (<)
    <|> op ">"   (>)
    )
    >>= return .fst
  where  op :: String -> (Int -> Int -> Bool) -> SqrParser (String, (Int -> Int -> Bool))
         op s o = try (string s *> return (s, o))

pTextOp = lexeme (
        op "||"  (++)
    )
    >>= return .fst
  where op s o = try (string s *> return (s, o))

pBitOp = lexeme (
        op "And"   (&&)
    <|> op "Or"    (||)
    <|> op "Xor"   (xor)
    )
    >>= return .fst
  where  op :: String -> (Bool -> Bool -> Bool) -> SqrParser (String, (Bool -> Bool -> Bool))
         op s o = try (stringCI s *> return (s, o))
         xor a b = (a || b) && not (a && b)

commandList = fmap stringCI [
    "ADD",
    "ALTER-COLOR-MAP",
    "ALTER-CONNECTION",
    "ALTER-LOCALE",
    "ALTER-PRINTER",
    "ALTER-REPORT",
    "ARRAY-ADD",
    "ARRAY-DIVIDE",
    "ARRAY-MULTIPLY",
    "ARRAY-SUBTRACT",
    "ASK",

    "BEGIN-DOCUMENT",
    "BEGIN-EXECUTE",
    "BEGIN-SELECT",
    "BEGIN-SQL",

    "BOX",
    "BREAK",
    "CALL",
    "CLEAR-ARRAY",
    "CLOSE",
    "COLUMNS",
    "COMMIT",
    "CONCAT",
    "CONNECT",
    "CREATE-ARRAY",
    "CREATE-COLOR-PALETTE",
    "DATE-TIME",
    "DECLARE PRINTER",
    "DECLARE PROCEDURE",
    "DISPLAY",
    "DIVIDE",
    "DO",
    "DOLLAR-SYMBOL",
    "ELSE",
    "ENCODE",

    "END-DECLARE",
    "END-DOCUMENT",
    "END-EVALUATE",
    "END-FOOTING",
    "END-HEADING",
    "END-IF",
    "END-PROCEDURE",
    "END-PROGRAM",
    "END-REPORT",
    "END-SELECT",
    "END-SETUP",
    "END-SQL",
    "END-WHILE",

    "EVALUATE",
    "EXECUTE",
    "EXECUTE",
    "EXIT-SELECT",
    "EXTRACT",
    "FIND",
    "FONT",
    "GET",
    "GET-COLOR",
    "GOTO",
    "GRAPHIC",
    "HORZ-LINE",
    "IF",
    "INPUT",
    "LAST-PAGE",
    "LET",
    "LOAD-LOOKUP",
    "LOOKUP",
    "LOWERCASE",
    "MBTOSBS",
    "MONEY-SYMBOL",
    "MOVE",
    "MULTIPLY",
    "NEW-PAGE",
    "NEW-REPORT",
    "NEXT-COLUMN",
    "NEXT-LISTING",
    "NO-FORMFEED",
    "OPEN",
    "PAGE-NUMBER",
    "PAGE-SIZE",
    "POSITION",
    "PRINT",
    "PRINT-BAR-CODE",
    "PRINT-CHART",
    "PRINT-DIRECT",
    "PRINTER",
    "PRINTER-DEINIT",
    "PRINTER-INIT",
    "PRINT-IMAGE",
    "PUT",
    "READ",
    "ROLLBACK",
    "SBTOMBS",
    "SECURITY",
    "SET-COLOR",
    "SET-GENERATIONS",
    "SET-LEVELS",
    "SET-MEMBERS",
    "SHOW",
    "STOP",
    "STRING",
    "SUBTRACT",
    "SYSTEM",
    "TOC-ENTRY",
    "UNSTRING",
    "UPPERCASE",
    "USE",
    "USE-COLUMN",
    "USE-PRINTER-TYPE",
    "USE-PROCEDURE",
    "USE-REPORT",
    "WHILE",
    "WRITE"
  ]

-----------------------------------------------------------
-- Numbers
-----------------------------------------------------------
-- integers and naturals
int             = do{ f <- lexeme sign
                    ; n <- nat
                    ; return (f n)
                    }

sign            =   (char '-' >> return negate)
                <|> (char '+' >> return id)
                <|> return id

nat             = zeroNumber <|> decimal

zeroNumber      = do{ char '0'
                    ; hexadecimal <|> decimal <|> return 0
                    }

decimal         = number 10 digit
hexadecimal     = do{ oneOf "xX"; number 16 hexDigit }

number radix baseDigit
    = do{ digits <- many1 baseDigit
        ; let n = Data.List.foldl (\x d -> radix * x + toInteger (digitToInt d)) 0 digits
        ; seq n (return n)
        }


naturalOrFloat  = lexeme (signedNatFloat) <?> "number"

float           = lexeme floating   <?> "float"
integer         = lexeme int        <?> "integer"
natural         = lexeme nat        <?> "natural"


-- floats
floating        = do{ n <- decimal
                    ; fractExponent n
                    }

signedNatFloat  = do{ f <- lexeme sign
                    ; n <- unsignedNatFloat
                    ; return (f n)
                    }

unsignedNatFloat =  do{ char '0'
                      ; unsignedNatFloat
                      }
                    <|> decimalFloat
                    <|> fractFloat 0

decimalFloat    = do{ n <- decimal
                    ; option (fromInteger n) (fractFloat n)
                    }

fractFloat n    = do{ f <- fractExponent n
                    ; return f
                    }

fractExponent n = do{ fract <- fraction
                    ; expo  <- option 1.0 exponent'
                    ; return ((fromInteger n + fract)*expo)
                    }
                <|>
                  do{ expo <- exponent'
                    ; return ((fromInteger n)*expo)
                    }

fraction        = do{ char '.'
                    ; digits <- many1 digit <?> "fraction"
                    ; return (Data.List.foldr op 0.0 digits)
                    }
                  <?> "fraction"
                where
                  op d f    = (f + fromIntegral (digitToInt d))/10.0

exponent'       = do{ oneOf "eE"
                    ; f <- sign
                    ; e <- decimal <?> "exponent"
                    ; return (power (f e))
                    }
                  <?> "exponent"
                where
                   power e  | e < 0      = 1.0/power(-e)
                            | otherwise  = fromInteger (10^e)


--setBuiltInVariables =

--defineVariable :: VarType -> VarIdentifier -> SqrParser ()

--define
