module Main (main) where

import System.Console.GetOpt
import System.Environment
import System.IO
import System.Exit
import System.FilePath
import System.Directory

import Data.List
import qualified Data.Map as M

import Control.Monad
import Text.Parsec
import Text.Parsec.Pos
import SqrppParser
import SqrParser
import SqrFile
import VarFileParser

data Options = Options  { optSearchPaths          :: [FilePath]
                        , optOnlyPreprocess       :: Bool
                        , optRecursiveCheck       :: Bool
                        , optFileFilter           :: FileFilter
                        , optOutputHandle         :: Handle
                        , optExceptCheckParam     :: Bool
                        , optOutputFormatList     :: Bool
                        , optVarFilePath          :: Maybe FilePath
                        , variablesSuppressError  :: [(Scope, VarIdentifier)]
                        , optSuppressOutput       :: Bool
                        }

type FileFilter = String -> Bool

startOptions :: Options
startOptions = Options  { optSearchPaths          = ["."]
                        , optOnlyPreprocess       = False
                        , optRecursiveCheck       = False
                        , optOutputHandle         = stdout
                        , optFileFilter           = (const True)
                        , optExceptCheckParam     = False
                        , optOutputFormatList     = False
                        , optVarFilePath          = Nothing
                        , variablesSuppressError  = []
                        , optSuppressOutput       = False
                        }

options :: [ OptDescr (Options -> IO Options) ]
options =
  [
      Option "Ii" []
        (ReqArg
            (\arg opt -> return opt { optSearchPaths = (splitSearchPath arg) })
            "<list>")
        ("Specify the search paths for include files.\n"
        ++ "Paths must be delimited by ';'(semicolon).\n"
        ++ "If want to include a current directory to \n"
        ++ "search paths, append a dot to the path list.\n\n"
        )

    , Option "v" []
        (NoArg
            (\_ -> do
                hPutStrLn stderr "sqrpp version 1.06"
                exitSuccess))
        "Display version.\n\n"

    , Option "Hh" []
        (NoArg
            (\_ -> do {showHelp; exitSuccess})
        )
        "Display usage.\n\n"

    , Option "Ee" []
        (NoArg (\opt -> return opt { optOnlyPreprocess=True } ) )
        "Output preprocessed source code.\n\n"

    , Option "Oo" []
        (ReqArg (\arg opt -> do
                      h <- openFile arg WriteMode
                      return opt { optOutputHandle=h } ) "<path>")
        ("Specify output file path.\n"
        ++ "When not specify, output to standard output.\n\n"
        )

    , Option "Rr" []
        (OptArg (\arg opt -> do { let opt1 = opt  { optRecursiveCheck=True }
                                ; return $ case arg of
                                            Nothing     -> opt1
                                            Just filter -> opt1 { optFileFilter=(matchFileName filter) }
                                } ) "<filter>" )
        ("Check recursively for included files.\n"
        ++ "If specify a filter, check only files matched to the filter.\n"
        ++ "When not specify a filter, check all included files.\n\n"
        )

    , Option "Uu" []
        (NoArg (\opt -> return opt { optExceptCheckParam=True } ) )
        ("suppress definitions error for procedure's parameters.\n\n")


    , Option "Ll" []
        (NoArg (\opt -> return opt { optOutputFormatList=True } ) )
        ("output with list format.\n\n")

    , Option "Pp" []
        (ReqArg (\arg opt -> return opt { optVarFilePath=Just arg } ) "<path>")
        (  "Specify a variable file path.\n"
        ++ "Variables on the file are suppressed definition error.\n"
        ++ "Following is an example of a variable file.\n\n"
        ++ "    #CURR_SEQ_NO\n"
        ++ "    $CURRENCY_CD\n"
        ++ "    get_state_code $COUNTRY_CD -- local variable with procedure name.\n"
        ++ "    get_state_code $STATE_CD\n"
        ++ "    get_state_code #ERROR_NO\n"
        ++ "\n")

    , Option "!" []
        (NoArg (\opt -> return opt { optSuppressOutput=True } ) )
        ("Suppress output.\n")
  ]


showHelp :: IO()
showHelp = do
    prg <- getProgName
    ln "Preprocess and report undefined variable for SQR file."
    ln $ "\n\t" ++ prg ++ " {filepath | [Options]}"
    ln (usageInfo "\nOptions" options)
    ln "(example)"
    ln "SQRPP gpjpye01 -I.;c:\\pt851\\sqr -E"
    ln "SQRPP -Ic:\\pt851\\sqr;. -Rgp* gpjpye01 -U -Pvar.txt"
  where ln = putStrLn



main :: IO()
main = do
    args <- getArgs

    -- Parse options, getting a list of option actions
    let (actions, nonOptions, errors) = getOpt Permute options args
    opts <- foldl (>>=) (return startOptions) actions

    case nonOptions of
      []       -> do{showHelp; exitFailure}
      (arg:xs) -> do  { filePath <- getSourceFilePath arg (optSearchPaths opts)
                      ; vars <- maybe (return []) getVariablesSuppressError (optVarFilePath opts)
                      ; let fileName  = takeFileName filePath
                            state     = initialState  { rootSourceName=fileName
                                                      , searchPaths=(optSearchPaths opts)
                                                      , matchSourceNameForReport=(optFileFilter opts)
                                                      , recursiveCheck=(optRecursiveCheck opts)
                                                      }
                            handle    = optOutputHandle opts
{-
                            doParse p f g = readFile filePath >>=
                                            runParserT (pSqr p) state fileName >>=
                                            (hPutStrLn handle) . (either f g)
-}
                            doParse p f g = do
                                        s <- readFile filePath
                                        r <- runParserT (pSqr p) state fileName s
                                        case r of
                                              Left err -> do  { myPrint (f err)
                                                              ; hClose handle
                                                              ; exitFailure
                                                              }
                                              Right x  -> do  { myPrint (g x)
--                                                              ; putStrLn (show vars)
                                                              ; hClose handle
                                                              }
                            myPrint s = if (optSuppressOutput opts) then (return ()) else (hPutStrLn handle s)

                            opts2 = opts {variablesSuppressError=vars}

                      ; if (optOnlyPreprocess opts) then (doParse pSqrPP      show id)
                                                    else (doParse pSqrProgram show (showError opts2) )
                      }
{-
                        if (optOnlyPreprocess opts) then (doParse pSqrPrint      show id)
                                                    else (doParse pSqrProgram show showResult)
                                  else (doParse pSqrProgram show showProcedures)
-}


getVariablesSuppressError varFilePath = do
    s <- readFile varFilePath
    let fileName = takeFileName varFilePath
    result <- runParserT pVarFile () fileName s
    case result of
        Left  x -> putStrLn (show x) >> exitFailure
        Right y -> return y

getSourceFilePath arg searchPaths = do
    (exists, path) <- findFile arg
    if exists
      then return path
      else do { (exists2, path2) <- inSearchPaths findFile arg searchPaths
              ; if exists2
              ;   then return path2
              ;   else return arg
              }

findFile path = do
    exists <- doesFileExist path
    if exists
      then return (True, path)
      else do  { let path2 = path ++ ".sqr"
                ; exists2 <- doesFileExist path2
                ; if exists2
                ;   then return (True, path2)
                ;   else return (False, "")
                }

inSearchPaths finder fileName [] = return (False, "")
inSearchPaths finder fileName (searchPath:rest) = do
                          { let candidate = searchPath </> fileName
                          ; ret@(exists, path) <- finder candidate
                          ; if exists
                          ;   then return ret
                          ;   else inSearchPaths finder fileName rest
                          }


showError opts st =
    let smap = scopeMap st
        pmap = procMap st
        globalScope = maybe M.empty id (M.lookup Global smap)
    in Data.List.foldl (++) "" (fmap (\(key, vmap) -> showVarMapError opts globalScope key pmap vmap) (M.assocs smap))


showVarMapError :: Options -> VarMap -> Scope -> ProcedureMap -> VarMap -> String
showVarMapError opts globalScope scope pmap vmap
                = let (proc_nm, is_param) = case scope of
                                    Global -> ("", const False)
                                    Local proc_nm -> let procedure = (M.lookup proc_nm pmap)
                                                         vars = maybe [] (fmap varIdentifier . parameters) procedure
                                                      in (maybe "" procName procedure, (flip elem vars))
                      in Data.List.foldl (++) "" $ fmap (showVariableError opts globalScope scope proc_nm is_param) (M.assocs vmap)



showVariableError opts = if optOutputFormatList opts then showVariableError1 opts else showVariableError0 opts

showVariableError0 opts globalScope scope proc_nm is_param (var, attr) =
    case reportPos attr of
      Just pos -> let vars    = variablesSuppressError opts
                      key     = (scope, var)
                      excpt_p = optExceptCheckParam opts in
        (if (elem key vars) || (defined attr) || (excpt_p && is_param var) then "" else (err var pos " is not defined.") )
                    ++ (case usedCount attr of
                            0 -> warning var pos " is not used."
                            1 -> warning var pos " used only once."
                            _ -> ""
                       )
        ++ (M.foldl (\accum pos -> accum ++ warning var pos " Underscore may be unnecessary. The scope is not local.") "" (warnUnderbarPos attr))
        ++ if isSameAsGlobalVarName opts globalScope scope is_param var attr then (warning var pos " is same name as global variable.") else ""
      _ -> ""

  where warning = msg "Warning"
        err     = msg "Error"
        scope_nm = if scope == Global then "Global" else ("Local " ++ proc_nm)
        msg label var pos  msg  = label ++ ":\t" ++ (showVariableName var attr)++ "\t(" ++ scope_nm ++ ")\t" ++ msg ++ "\t" ++ showPosition pos ++ "\n"

isSameAsGlobalVarName opts globalScope scope is_param var attr =
          scope /= Global
       && not (defined attr)
       && not (is_param var)
       && is_same_name
       && not_exists_in_var_file
  where is_same_name = maybe False (const True) (M.lookup var globalScope)
        not_exists_in_var_file = notElem (scope, var) (variablesSuppressError opts)

showVariableError1 opts globalScope scope proc_nm is_param (var, attr) =
    case reportPos attr of
      Just pos -> let vars    = variablesSuppressError opts
                      key     = (scope, var)
                      excpt_p = optExceptCheckParam opts
                      err_msg = (if (elem key vars) || (defined attr) || (excpt_p && is_param var) then [] else ["Error:NotDefined"] )
                      wrn_msgs = (case usedCount attr of
                                          0 -> ["NotUsed"]
                                          1 -> ["UsedOnlyOnce"]
                                          _ -> [])
                                 ++ (if isSameAsGlobalVarName opts globalScope scope is_param var attr then ["SameNameGlobalVar"] else [])
                      msg = err_msg ++ (if null wrn_msgs then [] else [unwords ( "Warning:" : wrn_msgs )] )
                   in if null msg
                        then ""
                        else (intercalate "\t" [(scopeStr scope proc_nm), (showVariableName var attr), (sourceName pos),  (show $ sourceLine pos), (unwords msg)]) ++ "\n"
      _ -> ""
--  where is_same_global_name len_nm = length (varName var) > len_nm &&  scope /= Global && isSameAsGlobalVarName globalScope var



showProcedures st = unlines $ fmap ((showProcedure (scopeMap st)) . snd) (M.assocs (procMap st))


showProcedure scope_map proc =
                  let proc_name = procName proc
                      var_map = maybe M.empty id (M.lookup (procScope proc) scope_map)
                    in proc_name ++ "\t"
                  ++ (case procScope proc of
                        Global -> show Global
                        _      -> let parms = parameters proc
                                    in (if null parms then "LOCAL" else (showParameters var_map parms))
                     )
                  ++ "\t" ++ (showPosition $ procPosition proc) ++ "\n"
  where showParameters var_map parms = "(" ++ foldl (++) "" (intersperse ", " (fmap (showParameter var_map) parms)) ++ ")"


showParameter var_map param =
    let var_id = varIdentifier param
        var_attr = maybe (initialVarAttr var_id) id (M.lookup var_id var_map)
      in (if (callType param) == CallByReference then ":" else "") ++ (showVariableName var_id var_attr)

  where initialVarAttr var_id = VarAttr  { varType= if (varSymbol var_id == Sharp) then TEXT else defaultNumericVarType
                                      , caseSensitiveName=(varName var_id)
                                      , defined=False
                                      , varPosition=nullPosition
                                      , reportPos=Nothing
                                      , usedCount=0
                                      , warnUnderbarPos=M.empty
                                      }
        nullPosition = initialPos ""


showResult st = showScopeMap (scopeMap st) (procMap st)


showScopeMap :: ScopeMap -> ProcedureMap -> String
showScopeMap smap pmap = unlines (fmap (\(key, vmap) -> (show key) ++ (procInfo key) ++ "\n" ++ showVarMap vmap) (M.assocs smap))
  where procInfo key = case key of
                          Global -> ""
                          Local pname -> case (M.lookup pname pmap) of
                                              Just proc -> let pos=procPosition proc in "\t" ++ (sourceName pos) ++ "\t" ++ (show (sourceLine pos))
                                              Nothing -> ""


showVarMap :: VarMap -> String
showVarMap vmap = unlines $ fmap (('\t':) . showVariable) (M.assocs vmap)


showVariable (var, attr) = showVariableName var attr
                        ++ "\tposition: " ++ (showPosition $ varPosition attr)
                        ++ "\treportPos: " ++ maybe "Nothing" showPosition (reportPos attr)
                        ++ "\tusedCount: " ++ (show $ usedCount attr)
                        ++ (if (defined attr) then "\t\t*** defined ***" else "")


showVariableName varIdent varAttr = (show (varSymbol varIdent)) ++ (caseSensitiveName varAttr)


showPosition pos = "("
                ++ (sourceName pos)
                ++ "\tline "
                ++ (show $ sourceLine pos)
--                ++ " column "
--                ++ (show $ sourceColumn pos)
                ++ ")"

scopeStr scope proc_nm = case scope of
                  Global -> "Global\t---"
                  Local _ -> "Local\t" ++ proc_nm
