{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
module SqrppParser(
  SqrParser,
  ParserState(..),
  initialState,
  pSqr,
  pSqrPrint,
  pSqrPP,
  eol,
  pFail,
  (+++),
  charCI,
  stringCI,
  nbsp,
  Scope(..),
  ScopeMap,
  VarName,
  VarMap,
  VarSymbol(..),
  VarType(..),
  VarIdentifier(..),
  VarAttr(..),
  Procedure(..),
  Parameter(..),
  CallType(..),
  StartWithUnderbar,
  fileLine,
  ProcedureMap,
  dbgPrint
) where
import Text.Parsec
import Text.Parsec.Pos
import Text.Parsec.Prim
import Text.Parsec.Error

import Text.Parsec.Language
import qualified Text.Parsec.Token as P

import Control.Applicative ((<*>), (*>), (<*), (<$>), liftA2)
import Control.Monad
import Control.Monad.Trans
import Data.Char
import qualified Data.Map as M
import Data.List
import Data.Maybe
import Data.Text (pack, unpack, strip)
import Debug.Trace
import System.Directory
import System.FilePath
import System.IO
import SqrFile
import ParserLib

dbgPrint s = do
    return ()
{-
    sinfo <- positionInfo
    liftIO (putStrLn $ s ++ "\t" ++ sinfo)
  where positionInfo :: SqrParser String
        positionInfo = do
            pos <- getPosition
            return $ (sourceName pos) ++ "\t" ++ (show (sourceLine pos))
-}

type SqrParser a = ParsecT String ParserState IO a

data Scope = Global | Local String
    deriving (Show, Eq, Ord)

type ScopeMap = M.Map Scope VarMap

type VarMap = M.Map VarIdentifier VarAttr

data VarSymbol = Dollar | Sharp
    deriving (Eq, Ord)

instance Show VarSymbol where
    show Sharp = "#"
    show Dollar = "$"

instance Read VarSymbol where
    readsPrec _ (x:xs) = case x of
                        '#' -> [(Sharp, xs)]
                        '$' -> [(Dollar, xs)]
                        _ -> []

data VarIdentifier = VarIdentifier  { varName   :: VarName
                                    , varSymbol :: VarSymbol
                                    }
    deriving (Show, Eq, Ord)

type VarName = String

type StartWithUnderbar = Bool

data VarAttr = VarAttr { varType            :: VarType
                       , caseSensitiveName  :: String
                       , defined            :: Bool
                       , usedCount          :: Int
                       , varPosition        :: SourcePos
                       , reportPos          :: Maybe SourcePos
                       , warnUnderbarPos    :: M.Map FileLine SourcePos
                       }
    deriving (Show, Eq)

type FileLine = (SourceName, Line)

fileLine :: SourcePos -> FileLine
fileLine pos = ((sourceName pos), (sourceLine pos))

data VarType = INTEGER | FLOAT | DECIMAL Int | TEXT | DATE
    deriving (Show, Eq, Ord)

type ProcedureMap = M.Map ProcName Procedure

data Procedure = Procedure { procName     :: ProcName
                           , procScope    :: Scope
                           , procPosition :: SourcePos
                           , parameters   :: [Parameter]
                           }
    deriving (Show, Eq, Ord)

type ProcName = String

data Parameter = Parameter  { varIdentifier :: VarIdentifier
                            , callType      :: CallType
                            }
    deriving (Show, Eq, Ord)

data CallType = CallByValue | CallByReference deriving (Show, Eq, Ord)

data ParserState = ParserState
                    { context :: [Context]
                    , symtab :: M.Map String String
                    , sequenceNo :: Int
                    , sourceNameStack :: [(Int, String, Int)]
                    , currentScope :: Scope
                    , scopeMap :: ScopeMap
                    , procMap :: ProcedureMap
                    , continued :: Bool
                    , bol :: Bool

                    , searchPaths :: [FilePath]
                    , matchSourceNameForReport :: String -> Bool
                    , recursiveCheck :: Bool
                    , rootSourceName :: String
                    }

initialState = ParserState  { context=[]
                            , symtab=M.empty
                            , sequenceNo=1
                            , sourceNameStack=[]
                            , currentScope=Global
                            , scopeMap = initialScopeMap
                            , procMap = M.empty
                            , continued = False
                            , bol = True
                            , searchPaths=[]
                            , recursiveCheck=False
                            , matchSourceNameForReport = const True
                            , rootSourceName = ""
                            }

initialScopeMap = M.insert Global M.empty M.empty

data Context = Context { ctxPos :: SourcePos
                       , ctxInp :: String }

(+++) :: SqrParser String -> SqrParser String -> SqrParser String
(+++) x y = (++) <$> x <*> y

infixl 1 +++

trim :: String -> String
trim = unpack . strip . pack

blank0 :: (Stream s m Char) => ParsecT s u m String
blank0 = many nbsp

blank1 :: (Stream s m Char) => ParsecT s u m String
blank1 = many1 nbsp

languageDef :: (Stream s m Char) => P.GenLanguageDef s ParserState m
languageDef = P.LanguageDef
               { P.commentStart   = ""
               , P.commentEnd     = ""
               , P.commentLine    = ""
               , P.nestedComments = False
               , P.identStart     = letter <|> char '_'
               , P.identLetter    = alphaNum <|> oneOf "_'"
               , P.opStart        = P.opLetter languageDef
               , P.opLetter       = oneOf ""
               , P.reservedOpNames= []
               , P.reservedNames   = []
               , P.caseSensitive  = False
               }

lexer :: (Stream s m Char) => P.GenTokenParser s ParserState m
lexer  = P.makeTokenParser languageDef

reserved :: (Stream s m Char) => String -> ParsecT s ParserState m ()
reserved  = P.reserved lexer

symbol :: (Stream s m Char) => String -> ParsecT s ParserState m String
symbol = P.symbol lexer

{-
main :: IO ()
main = let f = "a.sqr"
--main = let f = "c.sqr"
--main = let f = "gpjpit01.sqr"
--       in readFile f >>= runParserT (scan2 True) ([], Data.Map.empty) f >>= either (print) putStrLn
       in readFile f >>= runParserT (pSqr) initialState f >>= either (print) putStrLn
--         in readFile f >>= runParserT (ppNA_Line <* eof) ([], Data.Map.empty) f >>= either (print) (putStrLn)
-}


run :: Show a => ParsecT String ParserState IO a -> String -> IO ()
run p s = runParserT p initialState "" s >>= either (print) (print)


run2 = run ppNA_Line "#if 1=1\n\t#if 1=1\n\t\t#if 1=1\n\t\t#else\n\t\t#end-if\n\t#else\n\t#end-if\n#else\n#end-if\n"


ppSqr :: SqrParser String
ppSqr = do
      setBuiltSubstVar
--      st <- getState
--      liftIO (putStrLn (show (symtab st)))
      ppSqrLines <* eof

ppSqrLines :: SqrParser String
ppSqrLines = join <$> many1 pLine

eol0 :: SqrParser String
eol0 = do
    updateSequenceNo (1+)
    try $ string "\n" <|> string "\r\n" <|> restoreInput
    return "\n"

eol :: SqrParser String
eol = do
    string "\n"
    (oldSeqNo, newSeqNo) <- updateSequenceNo (1+)
--    liftIO (putStrLn (show oldSeqNo))
    ensureSoureName newSeqNo
    return "\n"



pSqr :: SqrParser a -> SqrParser a
pSqr parser = do  { ppSqr >>= setInput
                  ; updateSequenceNo (const 1)
                  ; getPosition >>= setPosition .initialPos . sourceName
                  ; st <- getState
                  ; let seqNo = sequenceNo st
                        newState = st {sourceNameStack=reverse (sourceNameStack st)}
                  ; setState newState
--                  ; liftIO (putStrLn (show (sourceNameStack newState)))
                  } >> parser

pSqrPrint :: SqrParser String
pSqrPrint = (join <$> many1 aLine) <* eof

pSqrPP :: SqrParser String
pSqrPP = many anyChar

aLine :: SqrParser String
aLine = do
  retVal <- rowNo +++ many (noneOf "\n\r") +++ eol
--  liftIO (putStrLn retVal)
  return retVal

rowNo :: SqrParser String
rowNo = do
    seqNo <- sequenceNo <$> getState
    pos <- getPosition
    let retVal = (show seqNo) ++ "\t"
               ++ (sourceName pos) ++ "\t"
               ++ (show (sourceLine pos)) ++ "\t"
    return retVal

ensureSoureName currentSeqNo = do
    st <- getState
    case (sourceNameStack st) of
       (n, f, m):xs ->  if n == currentSeqNo then
                            do  { setState (st {sourceNameStack=xs})
                            ; updateSourceNameAndLineNo f m
                            ; ensureSoureName currentSeqNo
                            }
                        else return ()
       _ -> return ()

updateSourceNameAndLineNo newSourceName newLineNo = do
      pos0 <- getPosition
      let pos1 = setSourceName pos0 newSourceName
          pos2 = setSourceLine pos1 newLineNo
      setPosition pos2

updateSequenceNo :: (Int -> Int) -> ParsecT String ParserState IO (Int, Int)
updateSequenceNo f = do
    st <- getState
    let oldSeqNo = sequenceNo st
        newSeqNo = f oldSeqNo
    setState st {sequenceNo = newSeqNo}
    return (oldSeqNo, newSeqNo)

-- pFail s = beginParserFail >> parserFail s
pFail s = do
        beginParserFail
        parserFail s

pLine :: SqrParser String
pLine = optional restoreInput >>
        (   pBlankLine
        <|> pBlankLine2
        <|> pCommentLine
        <|> do { sp <- blank0
--                ; body <- ppSetupBlock <|> pOtherSections <|> ppUnexpectedLine <|> ppDirectives <|> try (pTrailingComment >> eol0) <|> pCodeLine
                ; body <- ppSetupBlock <|> ppUnexpectedLine <|> ppDirectives <|> try (pTrailingComment >> eol0) <|> pCodeLine
                ; return (sp ++ body)
                }
        )


ppNA_Line :: SqrParser String
ppNA_Line = do
    optional restoreInput
    blank0 +++ (    pBlankLine
                    <|> pBlankLine2
                    <|> ppNA_IF
                    <|> ppUnexpectedElse
                    <|> ppUnexpectedEndIf
                    <|> (many1 (noneOf "\n\r") >> eol0)
                    )

ppNA_IF :: SqrParser String
ppNA_IF = do
        try $ do { stringCI "#if"
                 ; optional (optional (charCI 'n') >> stringCI "def")
                 ; ppEndWord
                 }
        many (noneOf "\n\r")
        eol0 +++ scan
  where
      scan =  optional restoreInput >> ((ppElseLine +++ scan2) <|> ppEndIfLine <|> (ppNA_Line +++ scan))
      scan2 = optional restoreInput >> (ppEndIfLine <|> (ppNA_Line +++ scan2))

pBlankLine :: SqrParser String
pBlankLine = try $ blank0 +++ (try (string "\n" <|> string "\r\n" <|> string "\r") <* updateSequenceNo (1+))

pBlankLine2 :: SqrParser String
pBlankLine2 = try $ blank1 +++ eol0

pCommentLine :: SqrParser String
pCommentLine = char '!' >> many (noneOf "\n\r") >> eol0

ppDirectives :: SqrParser String
ppDirectives = ppDefineLine <|> ppIfLine <|> ppIfDefLine <|> ppIfNDefLine <|> ppIncludeLine

ppUnexpectedLine :: SqrParser String
ppUnexpectedLine =  ppUnexpectedAskLine <|> ppUnexpectedEndSetupLine <|> ppUnexpectedEndIf <|> ppUnexpectedElse

ppUnexpectedEndIf :: SqrParser String
ppUnexpectedEndIf = (try ppEndIf) >> pFail "(SQR 4709) #ENDIF without preceding #IF."

ppUnexpectedElse :: SqrParser String
ppUnexpectedElse = (try ppElse) >> pFail "(SQR 4708) #ELSE without preceding #IF."

ppUnexpectedEndSetupLine :: SqrParser String
ppUnexpectedEndSetupLine = pEndSetupLine >> pFail "(SQR 4702) Line found outside paragraph.1"

ppUnexpectedAskLine :: SqrParser String
ppUnexpectedAskLine = (try ppAsk) >> pFail "(SQR 4702) Line found outside paragraph.2"


beginParserFail :: Monad m => ParsecT String ParserState m ()
beginParserFail = do
    optional (eof >> setInput " ")
    pos <- getPosition
    anyChar >> return ()
    setPosition pos
    return ()

ppDefineLine :: SqrParser String
ppDefineLine = do try (stringCI "#define" >> ppEndWord)
                  ret <- option "" body
                  ppEOL
                  return ret

  where body = try $ do
                  blank1
                  var <- trim <$> (try anyWord)
                  optional blank1
                  val <- trim <$> (option "" (join <$> many1 pCodeSegment))
                  if var /= "" then (saveSubstVar var val) else return ()
--                  liftIO (putStrLn $ var ++ "\t" ++ val)
                  return "\n"
--                  return $ var ++ ":\"" ++ val ++ "\"\n"

ppAsk :: SqrParser String
ppAsk = stringCI "ask" <* ppEndWord

ppAskLine :: SqrParser String
ppAskLine = do
    try (blank0 *> ppAsk)
    blank1
    varName <- anyWord
    prompt <- option ("Enter " ++ varName) (try (blank1 *> ppRightHandSide))
    ppEOL

    result <- lookupSubstVar varName
    case result of
        Nothing -> do { liftIO (putStr $ (trim prompt) ++ ": ")
                      ; liftIO (hFlush stdout)
                      ; value <- liftIO getLine
                      ; saveSubstVar varName value
                      ; liftIO (putStrLn $ (varName ++ ":" ++ value))
                      ; liftIO (hFlush stdout)
                      ; return "\n"
                      }
        Just v -> do {
                     ; liftIO (putStrLn $ ("`" ++ varName ++ "` is already defined. The value is `" ++ v ++ "`."))
                     ; return "\n"
                     }

ppIfLine :: SqrParser String
ppIfLine = do
    try ppIf
    blank1 <|> pFail "(SQR 4109) Invalid operator."

    line <- pCodeLineBody
    state <- getState
    pos <- getPosition
    inp <- getInput
    setInput line
    (condition, result) <- ppIFCondition
    setInput inp
    setPosition pos
    ppEOL
    body <- ppIfBody result
    return $ "\n" ++ body
--    return $ condition ++ "\n" ++ x

ppIfDefLine :: SqrParser String
ppIfDefLine  = try ppIfDef >> ppIfDefNDefBody True

ppIfNDefLine :: SqrParser String
ppIfNDefLine = try ppIfNDef >> ppIfDefNDefBody False

ppIfDefNDefBody :: Bool -> SqrParser String
ppIfDefNDefBody isIFDEF = do
    blank1
    line <- pCodeLineBody
    let varName = trim line
    result <- (maybe (not isIFDEF) (const isIFDEF)) <$> lookupSubstVar varName
    ppEOL
    body <- ppIfBody result
    return $ "\n" ++ body
--    return $ (if isIFDEF then "#IFDEF:" else "#IFNDEF:") ++ varName ++ ": " ++ (show result) ++ "\n" ++ x

ppIfBody result =  optional restoreInput >> (
                         (ppElseLine +++ elseBody result)
                     <|> ppEndIfLine
                     <|> (line result +++ ppIfBody result)
               )
    where elseBody result = optional restoreInput >> (ppEndIfLine <|> (line (not result) +++ elseBody result))
          line  result = if result then pLine else ppNA_Line

ppElseLine = do
    try (blank0 >> ppElse)
    many (noneOf "\n\r")
    eol0

ppEndIfLine = try $ do
    try (blank0 >> ppEndIf)
    many (noneOf "\n\r")
    eol0

ppIf :: SqrParser String
ppIf = stringCI "#if" <* ppEndWord

ppIfDef :: SqrParser String
ppIfDef = stringCI "#ifdef" <* ppEndWord

ppIfNDef :: SqrParser String
ppIfNDef = stringCI "#ifndef" <* ppEndWord

ppElse :: SqrParser String
ppElse = stringCI "#else" <* ppEndWord


ppSetupBlock :: SqrParser String
ppSetupBlock = do
    x <- try (stringCI "begin-setup" <* ppEndWord)
    dbgPrint ">> begin-setup"
    ppEOL
    y <- body
    return $ x ++ "\n" ++ y
  where body = do
              optional restoreInput
              (pEndSetupLine <* dbgPrint "<< end-setup") <|> ((++) <$> (try ppAskLine <|> pLine) <*> body)
{-
        printPos s = do
          pos <- getPosition
          liftIO (putStrLn $ (show (sourceLine pos)) ++ "\t" ++ s)
          return s
-}

pEndSetupLine :: SqrParser String
pEndSetupLine = try $ (skipMany nbsp >> stringCI "end-setup") +++ many (noneOf "\n\r") +++ eol0

ppEndIf :: SqrParser String
ppEndIf = do
    stringCI "#end"
    optional (char '-')
    stringCI "if" <* ppEndWord

ppEndWord = lookAhead (try ( (count 1 (satisfy isSpace)) <|> eol0 <|> (string "!" <* notFollowedBy (oneOf "!="))))

ppString :: SqrParser String
ppString = try ((char '\'') *> (many (noneOf "\'\n\r" <|> (try (char '\'' >> char '\'')))) <* (char '\''))


data PPValue = PPValue (String, (Maybe Double))
     deriving (Show)

instance Eq PPValue where
    PPValue (x, Nothing) == PPValue (y, Nothing) = x == y
    PPValue (_, Just x)  == PPValue (_, Just y)  = x == y
    PPValue (x, Just _)  == PPValue (y, Nothing) = x == y
    x == y = y == x

instance Ord PPValue where
    PPValue (x, Nothing) <= PPValue (y, Nothing) = x <= y
    PPValue (_, Just x)  <= PPValue (_, Just y)  = x <= y
    PPValue (x, Just _)  <= PPValue (y, Nothing) = x <= y
    x <= y = y >= x

ppEvalIfCond :: Monad m => (PPValue -> PPValue -> Bool) -> String -> String -> ParsecT String ParserState m Bool
ppEvalIfCond op lhs rhs = compare <$> (substitute lhs) <*> (substitute rhs)
    where substitute s = (maybe s id) <$> lookupSubstVar s
          compare x y = op (value x) (value y)
          value s  = PPValue ((fmap toUpper s), maybeRead s :: Maybe Double)

ppOperator :: Monad m => ParsecT String ParserState m (String, (PPValue -> PPValue -> Bool))
ppOperator = lexeme (   op "="   (==)
                    <|> op "!!=" (/=)
                    <|> op "!="  (/=)
                    <|> op "<>"  (/=)
                    <|> op "<="  (<=)
                    <|> op ">="  (>=)
                    <|> op "<"   (<)
                    <|> op ">"   (>)
                    )
                where op s o = try (string s *> return (s, o))

ppIFCondition :: ParsecT String ParserState IO (String, Bool)
ppIFCondition = do
    lhs <- trim <$> ppLeftHandSide
    blank0
    (opTxt, op) <- ppOperator <|> pFail "(SQR 4109) Invalid operator."
    blank0
    rhs <- trim <$> ppRightHandSide
    result <- ppEvalIfCond op lhs rhs
    return (("\"" ++ lhs ++ "\" " ++ opTxt ++ " \"" ++ rhs ++ "\" ---> " ++ show result), result)

ppLeftHandSide :: SqrParser String
ppLeftHandSide = ppString <|> (join <$> many (count 1 (noneOf "!\n\r\t =<>")))

ppRightHandSide :: SqrParser String
ppRightHandSide = ppString <|> (join <$> many1 pCodeSegment)

anyWord :: SqrParser String
anyWord = join <$> many1 (  pSubstitute
                        <|> (count 1 wordChar)
                        <|> pMinus
                        <|> (try $ string "!=")
                        <|> (try $ string "!!") )

wordChar :: Monad m => ParsecT String ParserState m Char
wordChar = noneOf "!\n\r\t -"

ppEOL :: SqrParser String
ppEOL = do
    optional pTrailingText
    eol0

pMinus :: SqrParser String
pMinus = try (string "-" <* notFollowedBy pTrailingText)

pSubstitute :: SqrParser String
pSubstitute = do
    pos <- getPosition
    var_name <- pSubstVar
    maybeVal <- lookupSubstVar var_name
    case maybeVal of
        Just val -> return val
        Nothing   -> do { putMessageIO pos ("(SQR 4707) No value found for substitution variable: {" ++ var_name ++ "}")
                        ; let val = "{" ++ var_name ++ "}"
                        ; saveSubstVar var_name val
                        ; return val
                        }
putMessageIO pos msg = do
      liftIO (putStrLn $ showPosition pos)
      liftIO (putStrLn msg)

showPosition pos = "\"" ++ (sourceName pos)
                ++ "\" (line "
                ++ (show $ sourceLine pos)
                ++ ", column "
                ++ (show $  sourceColumn pos) ++ "):"

pSubstVar :: SqrParser String
pSubstVar = try $ do
    char '{'
    x <- join <$> many1 (try (many1 $ noneOf "!\n\r\t }") <|> try (string "!=") <|> try (char '!' *> string "!"))
    char '}'
    return x

notSpace :: Monad m => ParsecT String ParserState m Char
notSpace = (satisfy (not . isSpace))

lexeme :: Monad m => ParsecT String ParserState m a -> ParsecT String ParserState m a
lexeme p = p <* blank0

pCodeLine :: SqrParser String
pCodeLine = pCodeLineBody +++ eol0

pCodeLineBody :: SqrParser String
pCodeLineBody = (join <$> many1 pCodeSegment) +++ option "" pTrailingText

pCodeSegment :: SqrParser String
pCodeSegment = anyWord +++ (pHyphen <|> blank0)

pHyphen :: SqrParser String
pHyphen = try $ (option " " blank1) +++ (string "-" <* lookAhead pTrailingText)



pOtherSections :: SqrParser String
pOtherSections =   pSection "DOCUMENT"
               <|> pSection "EXECUTE"
               <|> pSection "FOOTING"
               <|> pSection "HEADING"
               <|> pSection "REPORT"
               <|> pSection "SELECT"
               <|> pSection "SQL"
--               <|> pSection "SETUP"
--               <|> pSection "PROGRAM"
--               <|> pSection "PROCEDURE"

pSection :: String -> SqrParser String
pSection sectionName =  (startWith ("end-" ++ sectionName) >> pFail "(SQR 4702) Line found outside paragraph.3")
                    <|> (startWith ("begin-" ++ sectionName) +++ body)
    where body = endSectionLine <|> (pLine +++ body)
          endSectionLine = optional restoreInput >> keywordLine ("end-" ++ sectionName)

keywordLine :: String -> SqrParser String
keywordLine s = startWith s +++ many (noneOf "\n\r") +++ eol0

startWith :: String -> SqrParser String
startWith s = try ( (blank0 +++ stringCI s) <* ppEndWord )

keyword :: String -> SqrParser String
keyword s = try (stringCI s <* ppEndWord)

pTrailingText :: SqrParser String
pTrailingText = try $ do
    blank0
    optional pTrailingComment
    notFollowedBy (noneOf "\n\r")
    return ""

pTrailingComment :: SqrParser String
pTrailingComment = try $ do
            char '!' <* notFollowedBy (oneOf "!=")
            many $ noneOf "\n\r"

maybeRead s = do
     case (listToMaybe . reads) s of
        Just (x, "") -> Just x
        _ -> Nothing



ppIncludeLine :: SqrParser String
ppIncludeLine = do
    try (stringCI "#include" >> ppEndWord)
--    dbgPrint "#include start"
    optional blank1
    pos <- getPosition
    quote <- oneOf "\'\""
    f <- many (noneOf (quote:[]))
    char quote

    st <- getState
    (fileExists, path) <- liftIO (searchFile f (searchPaths st))
    ppEOL
--    dbgPrint "#include end"
    if (not fileExists)
      then fileNotExistsError pos f
      else do { let fileName = takeFileName path
              ; pushContext =<< Context <$> getPosition <*> getInput
              ; setPosition $ initialPos fileName
              ; setInput =<< liftIO (readFile path)
              ; st <- getState
              ; pos <- getPosition
              ; let seqNo = sequenceNo st
                    rowNo = sourceLine pos
              ; eof <|> setState (st {sourceNameStack=(seqNo,fileName,1):(sourceNameStack st)})
              }
    return "\n"

  where fileNotExistsError pos fileName = putMessageIO pos ("(SQR 4713) Cannot open the #INCLUDE file: '" ++ fileName ++ "'")

restoreInput :: SqrParser String
restoreInput = do
    eof
    state <- getState
    case context state of
        (x:xs) -> do { Context pos inp <- popContext
                     ; setPosition pos
                     ; setInput inp
                     ; st <- getState
                     ; let seqNo = sequenceNo st
                           rowNo = sourceLine pos
                           f = sourceName pos
                     ; pushSourceNm (seqNo, f,rowNo)
                     ; s <- option "" restoreInput
                     ; return $ "\n" ++ s
                     }
        _ -> return ""



pushSourceNm :: Monad m => (Int,String,Int) -> ParsecT s ParserState m ()
pushSourceNm x = do st <- getState
                    setState (st {sourceNameStack=x:(sourceNameStack st)})


popSourceNm :: Monad m => ParsecT s ParserState m ()
popSourceNm = do
  st <- getState
  case (sourceNameStack st) of
    x:xs -> setState (st {sourceNameStack=xs})
    _ -> return ()

pushContext :: Monad m => Context -> ParsecT s ParserState m ()
pushContext c = getState >>=  (\st -> putState (st {context= c:(context st)}))

popContext :: Monad m => ParsecT s ParserState m Context
popContext = do
    oldState@(ParserState {context=(x:xs)}) <- getState
    putState (oldState {context=xs})
    return x

saveSubstVar :: Monad m => String -> String -> ParsecT s ParserState m ()
saveSubstVar var val = do
    oldState <- getState
    let newSymtab = updateSymtab var val (symtab oldState)
    putState (oldState {symtab=newSymtab})

updateSymtab var val = M.insert (fmap toUpper var) val

builtInSubstVars =
  [ "sqr-database"
  , "sqr-dbcs"
  , "sqr-encoding"
  , "sqr-hostname"
  , "sqr-platform"
  ]

setBuiltSubstVar = do
    oldState <- getState
    let newSymtab = (Data.List.foldl (.) id (fmap builtin builtInSubstVars)) (symtab oldState)
    putState (oldState {symtab=newSymtab})
  where builtin substVarName = updateSymtab substVarName ("{" ++ substVarName ++ "}")

lookupSubstVar :: Monad m => String -> ParsecT s ParserState m (Maybe String)
lookupSubstVar var = getState >>= return . (M.lookup (fmap toUpper var)). symtab
